unit solve_unit;

{$MODE Delphi}

{procedures and data structures to solve hashi / bridges puzzles}

interface

uses LCLIntf, LCLType, LMessages;

function makebridgelist : boolean;
function searchsolution : byte;//0:end; 1:
procedure clearCC;
procedure getSolveBridgeData(var n,k : byte);
function getCCupdates : double;

implementation

uses game_unit,paint_unit;

const bridgeSpace = 50;  //maximum nr of bridge to add

type TBridge = record
                x1,y1,x2,y2 : byte;//game coords of islands
                AB : byte;         //0: no bridge; 1 : bridge
                Skip10 : boolean;  //skip double 01-10
               end;

var bridgelist : array[1..bridgespace] of Tbridge;
    maxbridge  : byte;
    CC : array[1..bridgeSpace] of byte; //combinations counter
    p  : byte;                          //digit pointer to CC
    BCcount : byte;                     //bridges (1's) set in CC
    BCmax : byte;
    CCupdates : double;

procedure getSolveBridgeData(var n,k : byte);
begin
 n := maxbridge;
 k := BCmax;
end;

function getCCupdates : double;
begin
 result := CCupdates;
end;

procedure clearbridgelist;
var i : byte;
begin
 for i := 1 to bridgespace do
  with bridgelist[i] do
   begin
    x1 := 0;
    y1 := 0;
    x2 := 0;
    y2 := 0;
    skip10 := false;
   end;
 maxbridge := 0;
end;

function makebridgelist : boolean;
//exit false if bridgelist >= 40
var a,b,i,j,n,nbr : byte;
    fbv : TVector;
begin
 result := true;
 clearbridgelist;
 n := 0;   //bridgelist entry
 for i := 1 to islandcount do
  begin
   b := island[i].brCount -  selectedbridgecount(i,$b);
   if b > 2 then b := 2;
   freebridgevector(fbv,i);
   a := fbv[1] - island[i].brConn[1];
   if a > b then a := b;
   nbr := island[i].neighbrs[1];
   for j := 1 to a do
    begin
     inc(n);
     bridgelist[n].x1 := island[i].x;
     bridgelist[n].y1 := island[i].y;
     bridgelist[n].x2 := island[nbr].x;
     bridgelist[n].y2 := island[nbr].y;
     bridgelist[n].skip10 := j=2;
     bridgelist[n].AB := 0;
    end;
   b := island[i].brCount - selectedbridgecount(i,$d);
   if b > 2 then b:= 2;
   a := fbv[4] - island[i].brConn[4];
   if a > b then a := b;
   nbr := island[i].neighbrs[4];
   for j := 1 to a do
    begin
     inc(n);
     bridgelist[n].x1 := island[i].x;
     bridgelist[n].y1 := island[i].y;
     bridgelist[n].x2 := island[nbr].x;
     bridgelist[n].y2 := island[nbr].y;
     bridgelist[n].skip10 := j=2;
     bridgelist[n].AB := 0;
    end;
   if n >= 40 then begin
                    result := false;
                    exit;
                   end;
  end;//for i
 maxbridge := n;
end;

procedure clearCC;
var i : byte;
begin
 for i := 1 to maxbridge do CC[i] := 0;
 p := 0;
 BCcount := 0;
 BCmax := maxbridgeCount - bridgecount;//bridges to place <----game_unit
 CCupdates := 0;
end;

function advanceCC : byte;
//0:end; 1:new combination
label loop,Ladd,Lshift,Lback;
begin

Loop:

 if p = maxbridge then goto Lback;
 if BCcount = BCmax then goto Lshift;

Ladd:

 inc(p);
 CC[p] := 1;
 inc(BCcount);
 if BCcount = BCmax then begin
                          result := 1;//report new combination
                          CCupdates := CCupdates + 1;
                          exit;
                         end;
 goto loop;

Lshift:

  CC[p] := 0;
  inc(p);
  CC[p] := 1;
  result := 1;
  CCupdates := CCupdates + 1;
  exit;                               //report new combination

Lback:

  CC[p] := 0;
  dec(BCcount);
  while (p > 0) and (CC[p] = 0) do dec(p);
  if p = 0 then begin
                 result := 0;
                 exit;                //report end
                end;
  CC[p] := 0;
  inc(p);
  CC[p] := 1;
  if (maxbridge-p) + BCcount < BCmax then goto LBack;
  goto Loop;
end;

function test01 : boolean;
//searc CC for double bridges and SKIP flag
//return false if CC skip condition
var i : byte;
begin
 result := true;
 for i := 2 to maxbridge do
  if (CC[i]=1) and (CC[i-1]=0) and bridgelist[i].Skip10 then
   begin
    result := false;
    exit;
   end;
end;

function procCC : byte;
//add/delete bridges acc to CC
//0: errror; 1:OK
var i : byte;
    line : TLine;
    OK : boolean;
begin
 result := 1;
 for i := 1 to maxbridge do
  if (CC[i] = 0) and (bridgelist[i].AB = 1) then
   begin
    line.x1 := bridgelist[i].x1;
    line.y1 := bridgelist[i].y1;
    line.x2 := bridgelist[i].x2;
    line.y2 := bridgelist[i].y2;
    deleteBridge(line);
    bridgelist[i].AB := 0;
   end;
 for i := 1 to maxbridge do
  if (CC[i] = 1) and (bridgelist[i].AB = 0) then
   begin
    line.x1 := bridgelist[i].x1;
    line.y1 := bridgelist[i].y1;
    line.x2 := bridgelist[i].x2;
    line.y2 := bridgelist[i].y2;
    OK := AddBridge(line);
    if OK = false then begin
                        result := 0;
                        exit;
                       end;
    bridgelist[i].AB := 1;
   end;
end;

function searchsolution : byte;
//return 0 if end; 1 if solution; 2 : timeout
var SL : boolean;
    timeout : dword;
begin
 result := 0;
 if maxbridge = 0 then exit;

 timeout := 1000;
 SL := false;
 repeat
  if AdvanceCC = 0 then exit;//report end
//
  if test01 then
   if procCC = 1 then
    SL := checkgamesolved;
  dec(timeout);
 until SL or (timeout = 0);
 if SL then result := 1 else result := 2;
end;

end.
