unit game_unit;

{
 data and procedures for hashi puzzle
 max game size 15 * 15
}

interface

uses paint_unit;

type  TFieldType = (ftNone,ftIsland,ftHbridge,ftVbridge);
      TVector = array[1..4] of byte;//[1] right; [2] up; [3] left; [4] down
      TIsland = record
                 x,y      : byte;    //coords in game
                 brCount  : byte;    //nr of bridges to install
                 brConn   : Tvector; //bridges connected per direction
                 neighbrs : Tvector; //index of neighbour if bridge present
                 groupNr  : byte;    //for isolation testing
                end;
      TField = record
                ftype    : TFieldType;
                Idx      : byte;     //entry in island list
                brCount  : byte;
               end;
      Twarning = record
                  wtype : byte;//1;cannot place all bridges; 2:isolated group
                  Nr : byte;   //1: island index; 2 : groupNr
                 end;
         THint = record
                  valid : boolean;
                  brs : byte;// nr of bridges
                  isNR : byte; //hint island index
                  dir : byte; //direction 1,2,3,4
                 end;

const maxgame = 15;
      maxIsland = 100;

var game : array[1..maxgame,1..maxgame] of TField;
    gamesize : byte = 7;
    Island : array[1..maxIsland] of Tisland;
    IslandCount : byte; //total islands
    BridgeCount : byte; //total bridges added in game
    maxBridgeCount : byte;

procedure clearGame;
procedure clearIslands;
procedure clearbridges;
function buildIslandlist : boolean;
function addbridge(const line : Tline) : boolean;
function deletebridge(const line : Tline) : boolean;
function checkGameSolved : boolean;
function testwarnings(var warning : Twarning) : boolean;
function testHints(var hint : Thint) : boolean;
function freeroad(n1,n2 : byte) : boolean;
function SelectedBridgeCount(n,mask : byte) : byte;
procedure freeBridgeVector(var fbv : TVector; n : byte);

implementation

uses unit1;

const maxgroup = 100;

type TGroupData = record
                    IsGCount : byte;    //islands count
                    BrSum : byte;       //bridges summed
                    BrConnected : byte; //actual bridges
                   end;

var grouplist : array[1..maxgroup] of TGroupData;

procedure clearGame;
var i,j : byte;
begin
 for j := 1 to maxgame do
  for i := 1 to maxgame do
   with game[i,j] do
    begin
     ftype := ftNone;
     brCount := 0;
     Idx := 0;
    end;
 islandcount := 0;
 bridgecount := 0;
 maxbridgecount := 0;
end;

procedure clearIslands;
var w : word;
    i : byte;
begin
 for w := 1 to maxIsland do
  with Island[w] do
   begin
    x := 0;
    y := 0;
    brCount := 0;
    for i := 1 to 4 do brConn[i] := 0;
    for i := 1 to 4 do neighbrs[i] := 0;
    groupNr := 0;
   end;
 islandcount := 0;
 bridgecount := 0;
 maxbridgecount := 0;
end;

procedure clearbridges;
//remove all bridges in game[ ]
var i,j : byte;
begin
 for j := 1 to gamesize do
  for i := 1 to gamesize do
   with game[i,j] do
    if (ftype = ftHbridge) or (ftype = ftVbridge) then
     begin
      brCount := 0; ftype := ftNone;
     end;
 for i:= 1 to islandCount do
  for j := 1 to 4 do island[i].brConn[j] := 0;
 buildIslandList;
end;

procedure BridgeVector(var bv : Tvector;x,y : byte);
//from game[ ]
var i : byte;
begin
 for i := 1 to 4 do bv[i] := 0;
 for i := 1 to 4 do
  case i of
   1 : if (x < gamesize - 1) then
        with game[x+1,y] do
         if ftype = ftHbridge then bv[i] := brCount;
   2 : if (y > 2) then
        with game[x,y-1] do
         if ftype = ftVbridge then bv[i] := brCount;
   3 : if (x > 2) then
        with game[x-1,y] do
         if ftype = ftHbridge then bv[i] := brCount;
   4 : if (y < gamesize - 1) then
        with game[x,y+1] do
         if ftype = ftVbridge then bv[i] := brCount;
  end;
end;

procedure GameRightNeighbour(var nb : byte; x,y : byte);
//called if possible neighbour
//ignore bridges
//check game[ ]
begin
 repeat
  inc(x);
  with game[x,y] do
   if ftype = ftIsland then nb := Idx;
  until (nb > 0) or (x = gamesize);
end;

procedure GameUpNeighbour(var nb : byte; x,y : byte);
//called if possible neighbour
//ignore bridges
//check game[ ]
begin
 repeat
  dec(y);
  with game[x,y] do
   if ftype = ftIsland then nb := Idx;
  until (nb > 0) or (y = 1);
end;

procedure GameLeftNeighbour(var nb : byte; x,y : byte);
//called if possible neighbour
//ignore bridges
//check game[ ]
begin
 repeat
  dec(x);
  with game[x,y] do
   if ftype = ftIsland then nb := Idx;
  until (nb > 0) or (x = 1);

end;

procedure GameDownNeighbour(var nb : byte; x,y : byte);
//called if possible neighbour
//ignore bridges
//check game[ ]
begin
 repeat
  inc(y);
  with game[x,y] do
   if ftype = ftIsland then nb := Idx;
  until (nb > 0) or (y = gamesize);
end;

procedure getConnectVector(var cv : TVector; n : byte);
//n is island list index
//by bridge connected islands
var i : byte;
begin
 for i := 1 to 4 do cv[i] := 0;
 for i := 1 to 4 do
  with island[n] do
   if brConn[i] > 0 then cv[i] := neighbrs[i];
end;

function scalarsum(const v : Tvector) : byte;
var i : byte;
begin
 result := 0;
 for i := 1 to 4 do result := result + v[i];
end;

function buildIslandlist : boolean;
//also set idx in game[ ]
var i,j,n : byte;
begin
 clearIslands;
 n := 0;
 for j := 1 to gamesize do
  for i := 1 to gamesize do
   with game[i,j] do
    case ftype of
     ftIsland : begin
                 inc(n); //index for island list
                 with island[n] do
                  begin
                   x := i;
                   y := j;
                   brCount := game[i,j].brCount;
                   maxbridgecount := maxbridgecount + brCount;
                   BridgeVector(brConn,i,j);
                   bridgecount := bridgecount + scalarsum(brConn);
                   groupNr := 0;
                  end;
                 Idx := n;//index into island list
                end;
    end;
 islandcount := n;
 result := (maxbridgecount and $1) = 0;
 maxbridgecount := maxbridgecount shr 1;//divide by 2
 bridgecount := bridgecount shr 1;
//
 for n := 1 to islandcount do      //find neighbours
  with island[n] do
   for i := 1 to 4 do
    case i of
     1 : if x < gamesize-1 then GameRightNeighbour(neighbrs[i],x,y);
     2 : if y > 2 then GameUpneighbour(neighbrs[i],x,y);
     3 : if x > 2 then GameLeftNeighbour(neighbrs[i],x,y);
     4 : if y < gamesize-1 then GameDownNeighbour(neighbrs[i],x,y);
    end;
end;

function SelectedBridgeCount(n,mask : byte) : byte;
//mask xxx1: right; xx1x: up; x1xx: left; 1xxx: down
var i : byte;
begin
 result := 0;
 for i := 1 to 4 do
  with island[n] do
   begin
    if (mask and $1) > 0 then result := result + brConn[i];
    mask := mask shr 1;
  end;
end;

function freeroad(n1,n2 : byte) : boolean;
//check no obstructions between islands n1, n2
var h,i,x1,y1,x2,y2 : byte;
    hor,vert : boolean;
begin
 result := false;
 x1 := island[n1].x; y1 := island[n1].y;
 x2 := island[n2].x; y2 := island[n2].y;
 if x1 > x2 then
  begin
   h := x1; x1 := x2; x2 := h;
  end;
 if y1 > y2 then
  begin
   h := y1; y1 := y2; y2 :=h;
  end;
 hor := y1 = y2;
 vert := x1 = x2;
 if hor then
  begin
   i := x1 + 1;
   repeat
    if game[i,y1].ftype = ftVbridge then exit;
    inc(i);
   until i = x2;
  end;
 if vert then
  begin
   i := y1;
   repeat
    if game[x1,i].ftype = ftHbridge then exit;
    inc(i);
   until i = y2;
  end;
 result := true;
end;

procedure freeBridgeVector(var fbv : TVector; n : byte);
//n is island nr
//fbv is nr of available bridges per direction 1..4
const mask : array[1..4] of byte = ($b,$7,$e,$d);
var OK : boolean;
    i,fnbc,nb : byte;
begin
 for i := 1 to 4 do
  begin
   fnbc := 0;
   nb := island[n].neighbrs[i];
   if nb > 0 then
    begin
     OK := (island[n].brConn[i] > 0) or freeroad(n,nb);
     if OK then
      begin
       fnbc := island[nb].brCount - selectedbridgecount(nb,mask[i]);
       if (island[n].brCount = 1) and (island[nb].brCount = 1) then fnbc := 0;
       if (island[n].brCount = 2) and (island[nb].brCount = 2) and (fnbc = 2) then
          fnbc := 1;
      end;
    end;
   if fnbc > 2 then fnbc := 2;
   fbv[i] := fnbc;
  end;
end;

function FreeBridgeCount(n : byte) : byte;
//return # bridges that may be connected to neighbours
const mask : array[1..4] of byte = ($b,$7,$e,$d);
var fbv : TVector;
begin
 freebridgevector(fbv,n);
 result := scalarsum(fbv);
end;

//-- add / delete bridges --

function addbridge(const line : Tline) : boolean;
//x1,y1    and    x2,y2 are island coordinates
//line coords must be sorted ..< ..
var hor : boolean;
    vert : boolean;
    bc,i,n1,n2 : byte;
begin
 result := false;
 with line do
  begin
   hor := x1 <> x2;
   vert := y1 <> y2;
   if (hor xor vert) = false then exit;
//
   n1 := game[x1,y1].Idx;
   n2 := game[x2,y2].Idx;
   bc := selectedbridgecount(n1,$f);
   bc := island[n1].brCount - bc;
   if bc = 0 then exit;
//
   bc := selectedbridgecount(n2,$f);
   bc := island[n2].brCount - bc;
   if bc = 0 then exit;
//
   if hor then
    if island[n1].brConn[1] = 2 then exit;
//
   if vert then
    if island[n1].brConn[4] = 2 then exit;
//
   if freeroad(n1,n2) = false then exit;
//
   if hor then
    begin
     for i := x1+1 to x2-1 do
      with game[i,y1] do
       begin
        inc(brCount);
        ftype := ftHbridge;
       end;
     inc(island[n1].brConn[1]);
     inc(island[n2].brConn[3]);
    end;
   if vert then
    begin
     for i := y1+1 to y2-1 do
      with game[x1,i] do
       begin
        inc(brCount);
        ftype := ftVbridge;
       end;
     inc(island[n1].brConn[4]);
     inc(island[n2].brConn[2]);
    end;
   inc(bridgecount);
   result := true;
  end;
end;

function deleteBridge(const line : Tline): boolean;
var hor : boolean;
    vert : boolean;
    i,nr1,nr2 : byte;
begin
 result := false;
 with line do
  begin
 hor := x1 <> x2;
 vert := y1 <> y2;
 if (hor xor vert) = false then exit;
 nr1 := game[x1,y1].Idx;
 nr2 := game[x2,y2].Idx;
//
 if hor then
  begin
   for i := x1+1 to x2-1 do
    with game[i,y1] do
     begin
      dec(brCount);
      if brCount = 0 then fType := ftNone;
     end;
   dec(island[nr1].brConn[1]);
   dec(island[nr2].brConn[3]);
  end;
//
 if vert then
  begin
   for i := y1+1 to y2-1 do
    with game[x1,i] do
     begin
      dec(brCount);
      if brCount = 0 then fType := ftNone;
     end;
   dec(island[nr1].brConn[4]);
   dec(island[nr2].brConn[2]);
  end;
 dec(bridgecount);
 result := true;
 end;
end;

//-- groups --

procedure setgroups;
//sets the group fields of islands
var i : byte;
    CHF : boolean;
    A,B,nb : byte;
begin
 for i := 1 to islandcount do island[i].groupNr := i;
 repeat
  CHF := false;          //change flag
  for i := 1 to islandcount do
   with island[i] do
    begin
     if brConn[1] > 0 then
      begin
       A := groupNr;
       nb := neighbrs[1];
       B := island[nb].groupNr;
       if A > B then island[i].groupNr := B
        else island[nb].groupNr := A;
        CHF := CHF or (A <> B);
      end;
     if brConn[4] > 0 then
       begin
        A := groupNr;
        nb := neighbrs[4];
        B := island[nb].groupNr;
        if A > B then island[i].groupNr := B
         else island[nb].groupNr := A;
        CHF := CHF or (A <> B);
       end;
     end;
 until CHF = false;
end;

procedure countgroups;
var i,br,grp : byte;
begin
 for i := 1 to maxgroup do
  with grouplist[i] do begin
                        IsGCount := 0;     //island group count
                        brsum := 0;
                        brConnected :=0;
                       end;
 for i := 1 to islandcount do
  with island[i] do
   begin
    grp := groupNr;
    br := brCount;
    with grouplist[grp] do
     begin
      inc(IsGcount);
      brSum := brSum + br;
      brConnected := brconnected + selectedbridgecount(i,$f);
     end;
   end;
end;

function GetdirectionVector(n : byte) : Tvector;
//get directionVector of island n
//return 1 if neighbour
var i,nb : byte;
begin
 with island[n] do
  for i := 1 to 4 do
   begin
    nb := neighbrs[i];
    if (nb > 0) and (freeroad(n,nb)) then result[i] := 1
    else result[i] := 0;
   end;
end;

function checkGameSolved : boolean;
begin
 result := false;
 setgroups;
 countgroups;
 with grouplist[1] do
  if (IsGCount = IslandCount) and (brSum  = brConnected) then
   result := true;
end;

function testHints(var hint : Thint) : boolean;
var A,B,SB,i,n,VAB : byte;
    av : Tvector;
    p : shortInt;
begin
 result := false;
 hint.valid := false;
 n := 0;
 while (islandcount > n) and (result = false) do
  begin
   inc(n);
   with island[n] do
    begin
     SB := scalarsum(brConn);
     if brCount > SB then
      begin
       freeBridgeVector(av,n);
        A := scalarsum(av);
       B := brCount;
       if A >= B then
        begin
         VAB := A - B;
         if VAB < 2 then
          begin
           i := 0;
           while (i < 4) and (result = false) do
            begin
             inc(i);
             p := av[i] - brConn[i] - VAB;
             if p < 0 then p := 0;
             if p > 0 then
              begin
               hint.dir := i;
               hint.brs := p;
               hint.isNr := n;
               result := true;
               hint.valid := true;
              end;
            end;//while i
          end;
        end;//if A > B
      end;//if BR count
    end;//with island
  end;//while
end;

function testWarnings(var warning : Twarning) : boolean;
var fbc,n : byte;
begin
 n := 0;
 result := false;
 warning.wType := 0;
 if islandcount = 0 then begin
                          warning.wtype := 0;
                          exit;
                         end;
 while (n < islandcount) and (result = false) do
  begin
   inc(n);
   fbc := FreeBridgeCount(n);
   if fbc < island[n].brCount then result := true;
  end;
 if result then
  with warning do
   begin
    wtype := 1;//cannot place all bridges
    Nr := n;
    exit;
   end;
//
 if checkGameSolved then exit;//does setgroups,countgroups;do not report solution
// setgroups;     //in island[ ] list   ========================
// countgroups;   //sum                 ========================

//
 n := 0;
 repeat
  inc(n);
  with grouplist[n] do
   if (IsGCount > 0) and (brSum = brConnected) then result := true;
 until (n = maxgroup) or result;
 if result then
  with warning do
   begin
    wtype := 2;
    Nr := n;
   end;
end;

end.
